# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import types
from abc import abstractmethod
from types import TracebackType
from typing import Any
from typing import cast
from typing import List
from typing import Optional
from typing import Type

from tenacity import before_sleep_log
from tenacity import retry
from tenacity import retry_if_exception_type
from tenacity import stop_after_attempt
from tenacity import wait_exponential
from vdk.api.job_input import IManagedConnection
from vdk.internal.builtin_plugins.connection.connection_hooks import (
    ConnectionHookSpecFactory,
)
from vdk.internal.builtin_plugins.connection.managed_cursor import ManagedCursor
from vdk.internal.builtin_plugins.connection.pep249.interfaces import PEP249Connection
from vdk.internal.builtin_plugins.run import job_input_error_classifier
from vdk.internal.core import errors
from vdk.internal.util.decorators import closing_noexcept_on_close

logger = logging.getLogger(__name__)


class ManagedConnectionBase(PEP249Connection, IManagedConnection):
    """
    Different database providers can subclass this class to provide raw connection (by implement _connect)

    The raw connection will be managed by the application.
    Optionally further customization is allowed by plugins that implement ConnectionHookSpec
    """

    def __init__(
        self,
        log: logging.Logger = logger,
        db_con: Optional[PEP249Connection] = None,
        connection_hook_spec_factory: ConnectionHookSpecFactory = None,
    ):
        """
        this constructor MUST be called by inheritors
        """
        if log:
            self._log = log
        else:
            self._log = logging.getLogger(__name__)
        self._is_db_con_open: bool = db_con is not None
        self._db_con: Optional[PEP249Connection] = db_con
        self._connection_hook_spec_factory = connection_hook_spec_factory

    def __getattr__(self, attr):
        """
        Dynamic interception and delegation of any (non-overridden) attribute access.
        In case an attribute is not explicitly managed (customized by overriding e.g. cursor()) -
        this attribute is looked up then the call is delegated, ensuring default behaviour success path.

        First, the non-managed attribute call is redirected to the wrapped native connection if attribute available,
        otherwise to the superclass if attribute is present.
        If the attribute is not specified by both the native connection nor the superclass, an AttributeError is raised.

        For more details on customizing attributes access, see PEP562.
        """
        # native connection
        if hasattr(self._db_con, attr):
            if isinstance(getattr(self._db_con, attr), types.MethodType):

                def method(*args, **kwargs):
                    return getattr(self._db_con, attr)(*args, **kwargs)

                return method
            return getattr(self._db_con, attr)
        # superclass
        if hasattr(super(), attr):
            if isinstance(getattr(super(), attr), types.MethodType):

                def method(*args, **kwargs):
                    return getattr(super(), attr)(*args, **kwargs)

                return method
            return getattr(super(), attr)
        raise AttributeError

    # Retry to connect on exception and backoff exponentially in
    # 30s, 1m, 2m, 4m
    @retry(
        stop=stop_after_attempt(5),
        wait=wait_exponential(multiplier=30, min=30, max=240),
        retry=retry_if_exception_type(Exception),
        before_sleep=before_sleep_log(logger, logging.DEBUG),
        reraise=True,
    )
    def connect(self) -> PEP249Connection:
        """
        :return: PEP249 Connection object (managed)
        """
        if not self._is_db_con_open:
            db_con = self._connect()
            self._log.debug(f"Established {str(db_con)}")
            self._is_db_con_open = True
            self._db_con = db_con

        return self

    # def get_managed_connection(self) -> PEP249Connection:
    #     return self

    def execute_query(self, query: str) -> List[List[Any]]:
        """
        Execute SQL query.
        """
        with closing_noexcept_on_close(self._cursor()) as cur:
            cur.execute(query)

            try:
                """
                1. According to PEP 249 fetchall() should throw an exception when there is no result set
                produced by cursor.execute() (say insert into table).
                2. In pyodbc the cursor.rowcount property is always -1 => can rely upon.
                3. In impyla there is a handy property cursor.has_result_set.
                But it is not in PEP 249 and not supported by pyodbc implementation
                4. The only solution found so far is to try/catch the fetchall() call
                and swallow the exception in very narrow set of cases.
                """
                # TODO support for fetchmany
                res = cur.fetchall()
            except Exception as e:
                res = None
                if str(e) in (
                    "No results.  Previous SQL was not a query.",  # message in pyodbc
                    "Trying to fetch results on an operation with no results.",  # message in impyla
                    "no results to fetch",  # psycopg: ProgrammingError: no results to fetch
                ):
                    self._log.debug(
                        "Fetching all results from query SUCCEEDED. Query does not produce results (e.g. DROP TABLE)."
                    )
                else:
                    if job_input_error_classifier.is_user_error(e):
                        blamee = errors.ResolvableBy.USER_ERROR
                    else:
                        blamee = errors.ResolvableBy.PLATFORM_ERROR
                    errors.log_and_rethrow(
                        blamee,
                        self._log,
                        what_happened="Fetching all results from query FAILED.",
                        why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(e),
                        consequences=errors.MSG_CONSEQUENCE_DELEGATING_TO_CALLER__LIKELY_EXECUTION_FAILURE,
                        countermeasures=errors.MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION,
                        exception=e,
                    )
            return cast(
                List[List[Any]], res
            )  # we return None in case of DML. This is not PEP249 compliant, but is more convenient

    def cursor(self, *args, **kwargs):
        if hasattr(self._db_con, "cursor"):
            return ManagedCursor(
                self._db_con.cursor(*args, **kwargs),
                self._log,
                self._connection_hook_spec_factory,
            )
        return super().cursor()

    def commit(self, *args, **kwargs):
        if hasattr(self._db_con, "commit"):
            return self._db_con.commit(*args, **kwargs)
        return super().commit()

    def rollback(self, *args, **kwargs):
        if hasattr(self._db_con, "rollback"):
            return self._db_con.rollback(*args, **kwargs)
        return super().rollback()

    def close(self, *args, **kwargs) -> None:
        """
        Close database connection. No op if it is already closed.
        """
        try:
            if self._is_db_con_open:
                self._log.debug(f"Closing database connection {self} ... ")
                self._is_db_con_open = False
                self._db_con.close()
                self._log.debug("Closing database connection SUCCEEDED.")
        except Exception as e:
            self._log.exception(
                "Closing database connection FAILED. No problem, I'm continuing as if nothing happened.",
                e,
            )

    @abstractmethod
    def _connect(self) -> PEP249Connection:
        """
        this method MUST be implemented by inheritors and should return PEP249 Connection object (unmanaged)
        """
        raise NotImplementedError

    def _cursor(self):
        self.connect()
        return self.cursor()

    # @abstractmethod # inherit optionally e.g. in case database does not support select 1 for checks
    def _is_connected(self) -> bool:
        if None is self._is_db_con_open:
            return False
        if False is self._is_db_con_open:
            return False
        try:
            self._cursor().execute(" select 1 -- Testing if connection is alive.")
            return True
        except Exception as e:
            self._log.debug(
                f"Connection {self} is disconnected ('select 1' returned {e})"
            )
        return False

    def __enter__(self) -> Any:
        return self

    def __exit__(
        self,
        exception_type: Optional[Type[BaseException]],
        exception_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:  # @UnusedVariable
        self.close()

    def __del__(self) -> None:
        self.close()

    def __str__(self) -> str:
        return "ManagedConnection[ isConnected:{} {} ]".format(
            self._is_db_con_open, self._db_con
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, PEP249Connection):
            return False
        return self._db_con == o or self._db_con == o._db_con

    def __hash__(self) -> int:
        return hash(self._db_con)



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pluggy

"""
The plugin project name. It is used in auto-discovery of the hooks and plugins.
If external module needs to recognize as a plugin module they need to specify this as an entry point.
"""
GROUP_NAME = "vdk.plugin.run"

"""
Decorator helper class for marking functions as hook specifications.

The hook spec can be configured using following variables:

If ``firstresult`` is ``True`` the 1:N hook call (N being the number of registered
hook implementation functions) will stop at I<=N when the I'th function
returns a non-``None`` result.

If ``historic`` is ``True`` calls to a hook will be memorized and replayed
on later registered plugins.
"""
hookspec = pluggy.HookspecMarker(GROUP_NAME)

"""
Decorator helper class for marking functions as hook implementations.
The method name must match one of those defined as hookspec
The method signature - the arguments need to be subset of arguments defined in hookspec

If any hookimpl errors with an exception no further callbacks are invoked
and the exception is packaged up and delivered to any wrappers
before being re-raised at the hook invocation point.

Plugin execution order can be configured in the decorator with following variables:

If ``optionalhook`` is ``True`` a missing matching hook specification will not result
in an error (by default it is an error if no matching spec is found).

If ``tryfirst`` is ``True`` this hook implementation will run as early as possible
in the chain of N hook implementations for a specification.

If ``trylast`` is ``True`` this hook implementation will run as late as possible
in the chain of N hook implementations.

If ``hookwrapper`` is ``True`` the hook implementations needs to execute exactly
one ``yield``.  The code before the ``yield`` is run early before any non-hookwrapper
function is run.  The code after the ``yield`` is run after all non-hookwrapper
function have run.  The ``yield`` receives a :py:class:`.callers._Result` object
representing the exception or result outcome of the inner calls (including other
hookwrapper calls).
Example:

@hookimpl(hookwrapper=True)
def func():
    before_other_hooks_execute()
    outcome = yield
    # outcome.excinfo may be None or a (cls, val, tb)
    # outcome.get_result() will raise exception if there was an exception before.
    after_other_hooks_executed(outcome.get_result())
    outcome.force_result(new_res)  # set new return result

For more details see https://pluggy.readthedocs.io

"""
hookimpl = pluggy.HookimplMarker(GROUP_NAME)



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import Callable

from vdk.api.job_input import IProperties


class PropertiesNotAvailable(IProperties):
    """
    Implementation of IProperties that will throw an error if user tries to access them.
    """

    def __init__(self, error_handler: Callable[[str], None]):
        self._error_handler = error_handler
        pass

    def get_property(self, name, default_value=None):  # @UnusedVariable
        self.tell_user("get_property")

    def get_all_properties(self):  # @UnusedVariable
        self.tell_user("get_all_properties")

    def set_all_properties(self, properties):  # @UnusedVariable
        self.tell_user("set_all_properties")

    def tell_user(self, methodname: str) -> None:
        self._error_handler(methodname)




RELEASE_VERSION='0.3.571259835'
BUILD_DATE='Thu Jun 23 12:36:17 UTC 2022'
BUILD_MACHINE_INFO='Linux runner-opyxurxq-project-28359933-concurrent-02cqf4 5.4.181-99.354.amzn2.x86_64 #1 SMP Wed Mar 2 18:50:46 UTC 2022 x86_64 GNU/Linux'
GITLAB_CI_JOB_ID='2630958610'
GIT_COMMIT_SHA='e0ddb7656de46c4b3695307013f204b691cd3d35'
GIT_BRANCH='main'



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import List

from vdk.internal.core.config import Configuration
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.util.utils import parse_config_sequence

PROPERTIES_DEFAULT_TYPE = "PROPERTIES_DEFAULT_TYPE"
PROPERTIES_WRITE_PREPROCESS_SEQUENCE = "PROPERTIES_WRITE_PREPROCESS_SEQUENCE"


class PropertiesConfiguration:
    def __init__(self, config: Configuration):
        self.__config = config

    def get_properties_default_type(self) -> str:
        return self.__config.get_value(PROPERTIES_DEFAULT_TYPE)

    def get_properties_write_preprocess_sequence(self) -> List[str]:
        return parse_config_sequence(
            self.__config, key=PROPERTIES_WRITE_PREPROCESS_SEQUENCE, sep=","
        )


def add_definitions(config_builder: ConfigurationBuilder):
    config_builder.add(
        key=PROPERTIES_DEFAULT_TYPE,
        default_value=None,
        description="Set the default properties type to use. "
        "Plugins can register different properties types. "
        "This option controls which is in use"
        "It can be left empty, in which case "
        "if there is only one type registered it will use it."
        "Or it will use one register with type 'default' ",
    )
    config_builder.add(
        key="PROPERTIES_WRITE_PREPROCESS_SEQUENCE",
        default_value=None,
        description="""A string of comma-separated property types.
            Those types are priorly registered in the IPropertiesRegistry, by
            mapping a factory for instantiating each IPropertiesServiceClient
            property type handler.
            This comma-separated string value indicates the sequence in which those
            IPropertiesServiceClient implementations `write_properties` method
            will be invoked. For example:
                   PROPERTIES_WRITE_PREPROCESS_SEQUENCE="a-prefixed-property,
                   replicated-property"
            would mean that the properties data stored would be first
            processed by the `a-prefixed-property` client, then by the
            `replicated-property` client, and finally would be stored by
            the default properties client.
            In case of pre-processing failure, the default client won't be invoked.
            """,
    )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import pathlib
import sys
from pathlib import Path
from typing import cast

from click import ClickException
from vdk.api.data_job import IStandaloneDataJob
from vdk.api.job_input import IJobInput
from vdk.api.plugin.core_hook_spec import CoreHookSpecs
from vdk.api.plugin.core_hook_spec import JobRunHookSpecs
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins import builtin_hook_impl
from vdk.internal.builtin_plugins.internal_hookspecs import InternalHookSpecs
from vdk.internal.builtin_plugins.run.data_job import DataJobDefaultHookImplPlugin
from vdk.internal.builtin_plugins.run.data_job import JobArguments
from vdk.internal.builtin_plugins.run.execution_tracking import ExecutionTrackingPlugin
from vdk.internal.builtin_plugins.run.file_based_step import TYPE_PYTHON
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.builtin_plugins.run.step import Step
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.context import CoreContext
from vdk.internal.core.statestore import StateStore
from vdk.internal.plugin.plugin import PluginRegistry


class NoOpStepDataJobHookImplPlugin(DataJobDefaultHookImplPlugin):
    """
    Implementation that overrides of the functionality of the DataJobDefaultHookImplPlugin
    so that it runs a single NoOp step
    """

    @hookimpl
    def initialize_job(self, context: JobContext):
        step = Step(
            name="NoOpStep",
            type="noop",
            runner_func=lambda _step, _job_input: True,  # Intentional noop
            file_path=Path(__file__),
            job_dir=context.job_directory,
        )
        context.step_builder.add_step(step)


class StandaloneDataJob(IStandaloneDataJob):
    """
    A contextmanager that executes all the necessary "setup" logic needed to return an initialised IJobInput
    object and finalize logic when the context is exited

    Sample usage::

        with StandaloneDataJob(datajob_directory) as job_input:
            #... use job_input object to interact with SuperCollider
    """

    def __init__(
        self,
        data_job_directory: pathlib.Path | None,
        name: str | None = None,
        job_args: dict | None = None,
        extra_plugins: list = [],
    ):
        if data_job_directory is None and name is None:
            raise ValueError(
                "Cannot initialize DataJob. "
                "At least one of data job directory or data job name need to be set. "
            )
        self._name = data_job_directory.name if name is None else name
        self._data_job_directory = data_job_directory
        if job_args is None:
            job_args = {}
        self._job_args = job_args

        self._plugin_registry = PluginRegistry()
        self._plugin_registry.add_hook_specs(InternalHookSpecs)
        for plugin in extra_plugins:
            self._plugin_registry.load_plugin_with_hooks_impl(plugin)
        self._plugin_registry.load_plugins_from_setuptools_entrypoints()
        self._plugin_registry.add_hook_specs(CoreHookSpecs)
        self._plugin_registry.load_plugin_with_hooks_impl(
            builtin_hook_impl, "core-plugin"
        )

        cast(CoreHookSpecs, self._plugin_registry.hook()).vdk_start.call_historic(
            kwargs=dict(plugin_registry=self._plugin_registry, command_line_args=[])
        )

        conf_builder = ConfigurationBuilder()
        cast(CoreHookSpecs, self._plugin_registry.hook()).vdk_configure(
            config_builder=conf_builder
        )
        configuration = conf_builder.build()

        core_context = CoreContext(self._plugin_registry, configuration, StateStore())
        core_context.plugin_registry.load_plugin_with_hooks_impl(
            ExecutionTrackingPlugin()
        )
        cast(CoreHookSpecs, self._plugin_registry.hook()).vdk_initialize(
            context=core_context
        )

        self._core_context = core_context
        self._job_context = None

    def __enter__(self) -> IJobInput:
        try:
            self._core_context.plugin_registry.load_plugin_with_hooks_impl(
                NoOpStepDataJobHookImplPlugin(), NoOpStepDataJobHookImplPlugin.__name__
            )

            from vdk.internal.builtin_plugins.templates.template_impl import (
                TemplatesImpl,
            )

            self._job_context = JobContext(
                name=self._name,
                job_directory=self._data_job_directory,
                core_context=self._core_context,
                job_args=JobArguments(self._job_args),
                templates=TemplatesImpl(
                    job_name=self._name, core_context=self._core_context
                ),
            )

            # We need to call both the initialize_job and run_job hooks to get a fully initialized JobInput
            cast(JobRunHookSpecs, self._plugin_registry.hook()).initialize_job(
                context=self._job_context
            )
            cast(JobRunHookSpecs, self._plugin_registry.hook()).run_job(
                context=self._job_context
            )

            return self._job_context.job_input
        except:
            # Ensure __exit__ gets called so the vdk_exception hook can be triggered
            self.__exit__(*sys.exc_info())

    def __exit__(self, exc_type, exc_value, exc_traceback):
        exit_code = 0
        if exc_type:
            cast(CoreHookSpecs, self._plugin_registry.hook()).vdk_exception(
                exception=exc_value
            )
            exit_code = (
                exc_value.exit_code if isinstance(exc_value, ClickException) else 1
            )
        else:
            cast(JobRunHookSpecs, self._plugin_registry.hook()).finalize_job(
                context=self._job_context
            )
        cast(CoreHookSpecs, self._plugin_registry.hook()).vdk_exit(
            context=self._core_context, exit_code=exit_code
        )
        # Ensure the exception can be processed normally by calling code
        return False


class StandaloneDataJobFactory:
    @staticmethod
    def create(
        data_job_directory: pathlib.Path | None,
        name: str | None = None,
        job_args: dict | None = None,
        extra_plugins: list = [],
    ) -> IStandaloneDataJob:
        """
        Arguments:
            data_job_directory: pathlib.Path
                The source code of the data job that will be executed.
            name: Optional[str]
                The name of the job.  If omitted will be inferred from the director name.
            job_args: Optional[dict]
                Allows for users to pass arguments to data job run.
                Data Job arguments are also used for parameter substitution in queries, see execute_query docstring.
            extra_plugins: Optional[list]
                List of extra plugins to register.  Mostly useful during testing
        """
        return StandaloneDataJob(
            data_job_directory=data_job_directory,
            name=name,
            job_args=job_args,
            extra_plugins=extra_plugins,
        )
